////////////////////////////////////////////////////////////////////////////////
#include "brick.h"
#include <stdio.h>
////////////////////////////////////////////////////////////////////////////////
Coord& Coord::operator +=(const Coord& rhs)
{
    i += rhs.i; j += rhs.j; k += rhs.k; return *this;
}

Coord& Coord::operator -=(const Coord& rhs)
{
    i -= rhs.i; j -= rhs.j; k -= rhs.k; return *this;
}
////////////////////////////////////////////////////////////////////////////////
void Block::set(const Coord& c)
{
    int ic(cellIndex(c));
    
    m[ic/UINTCELLS] |= 1<<(ic%UINTCELLS);
}

void Block::unSet(const Coord& c)
{
    int ic(cellIndex(c));

    m[ic/UINTCELLS] &= ~(1<<(ic%UINTCELLS));
}

uint Block::hit(int i,int j,int k) const
{
    int ic(cellIndex(i,j,k));

    return m[ic/UINTCELLS] & 1<<(ic%UINTCELLS);
}

bool Block::operator==(const Block& rhs) const
{
    int i;
    
    for(i=0;i<CSUINT;++i) if(m[i]!=rhs.m[i]) return false;
    
    return true;
}

void Block::clear()
{
    int i;
    
    for(i=0;i<CSUINT;++i) m[i] = 0;
}

void Block::dump() const
{
    int i, j, k;

    printf("\n");
    for(j=NJ-1;j>=0;--j)
    {
        for(k=0;k<NK;++k)
        {
            for(i=0;i<NI;++i) printf(" %d",hit(i,j,k)?1:0);
            printf("     ");
        }
        printf("\n");
    }
}
////////////////////////////////////////////////////////////////////////////////
void Brick::addCell(int i,int j,int k)
{
    if(cells.size()>=MXBC)
    {
        printf("too much brick-cells\n"); return;
    }

    Coord c(i,j,k);

    cells.push_back(c);

    // update bounding box
    if(cells.size()>1)
    {
        if(i<ijk_mn.i) ijk_mn.i = i; else if(i>ijk_mx.i) ijk_mx.i = i;
        if(j<ijk_mn.j) ijk_mn.j = j; else if(j>ijk_mx.j) ijk_mx.j = j;
        if(k<ijk_mn.k) ijk_mn.k = k; else if(k>ijk_mx.k) ijk_mx.k = k;
    }
    else
    {
        ijk_mn = c;
        ijk_mx = c;
    }
}

Block Brick::asBlock() const
{
    Block b;
    int i;

    b.clear(); for(i=0;i<cells.size();++i) b.set(cells[i]);

    return b;
}

void Brick::shift(const Coord& dijk)
{
    int i;

    for(i=0;i<cells.size();++i) cells[i] += dijk;
    ijk_mn += dijk;
    ijk_mx += dijk;
}

void Brick::shiftToOrigin()
{
    shift(Coord(-ijk_mn.i,-ijk_mn.j,-ijk_mn.k));
}

void Brick::swapBoundingBox()
{
    // called after rotations
    if(ijk_mx.i<ijk_mn.i) swap(ijk_mx.i,ijk_mn.i);
    if(ijk_mx.j<ijk_mn.j) swap(ijk_mx.j,ijk_mn.j);
    if(ijk_mx.k<ijk_mn.k) swap(ijk_mx.k,ijk_mn.k);
}

bool Brick::rangeCheck() const
{
    return ijk_mn.i >= 0 && ijk_mx.i < NI &&
           ijk_mn.j >= 0 && ijk_mx.j < NJ &&
           ijk_mn.k >= 0 && ijk_mx.k < NK;
}

float Brick::minCellIndex() const
{
    // used to define some ordering of bricks;
    int i, ic(CELLS);

    for(i=0;i<cells.size();++i) if(cellIndex(cells[i])<ic) ic = cellIndex(cells[i]);

    return float(ic);
}

void Brick::rotIJK()
{
    int i;

    for(i=0;i<cells.size();++i) cells[i].rotIJK();
    ijk_mn.rotIJK();
    ijk_mx.rotIJK();
}

void Brick::rotK()
{
    int i;

    for(i=0;i<cells.size();++i) cells[i].rotK();
    ijk_mn.rotK();
    ijk_mx.rotK();
    swapBoundingBox();
    shiftToOrigin();
}

void Brick::rotII()
{
    int i;

    for(i=0;i<cells.size();++i) cells[i].rotII();
    ijk_mn.rotII();
    ijk_mx.rotII();
    swapBoundingBox();
    shiftToOrigin();
}

void Brick::rotJJ()
{
    int i;

    for(i=0;i<cells.size();++i) cells[i].rotJJ();
    ijk_mn.rotJJ();
    ijk_mx.rotJJ();
    swapBoundingBox();
    shiftToOrigin();
}

void Brick::rotKK()
{
    int i;

    for(i=0;i<cells.size();++i) cells[i].rotKK();
    ijk_mn.rotKK();
    ijk_mx.rotKK();
    swapBoundingBox();
    shiftToOrigin();
}
////////////////////////////////////////////////////////////////////////////////
void BrickRot::init(int id,const Brick& b)
{
    Brick brick(b);
    Block block;
    Coord c;
    int i1, i2, i3;

    nCell = brick.cells.size();
    nRot = 0;
    brickId = id;
    bricks.resize(0);
    blocks.resize(0);
    nijk.resize(0);
    nShift.resize(0);
    nMove = 0;

    brick.shiftToOrigin();

    for(i1=0;i1<2;++i1)
    {
        for(i2=0;i2<4;++i2)
        {
            for(i3=0;i3<3;++i3)
            {
                if(brick.rangeCheck())
                {
                    ++nRot;
                    block = brick.asBlock();
                    if(find(block)<0)
                    {
                        bricks.push_back(brick);
                        blocks.push_back(block);
                        c.i = NI-brick.ijk_mx.i;
                        c.j = NJ-brick.ijk_mx.j;
                        c.k = NK-brick.ijk_mx.k;
                        nijk.push_back(c);
                        nShift.push_back(c.i*c.j*c.k);
                        nMove += c.i*c.j*c.k;
                    }
                }
                brick.rotIJK();
            }
            brick.rotK();
        }
        brick.rotII();
    }

    // surface of the brick
    brick = bricks[0];
    block = blocks[0];
    nSurf = nCell*6;
    for(i1=0;i1<nCell;++i1)
    {
        c = brick.cells[i1];
        if(c.i>0&& block.hit(c.i-1,c.j,c.k)) --nSurf;
        if(c.i<NI-1&& block.hit(c.i+1,c.j,c.k)) --nSurf;
        if(c.j>0&& block.hit(c.i,c.j-1,c.k)) --nSurf;
        if(c.j<NJ-1&& block.hit(c.i,c.j+1,c.k)) --nSurf;
        if(c.k>0&& block.hit(c.i,c.j,c.k-1)) --nSurf;
        if(c.k<NK-1&& block.hit(c.i,c.j,c.k+1)) --nSurf;
    }
}

bool BrickRot::qualify(float& val) const // higher values are better
{
    // game-symmetries will later be removed from that asymmetric brick with
    // the highest value;

    if(nRot==bricks.size()) // asymmetric brick
    {
        //val = -nRot;
        //val = nMove;
        //val = -nRot+0.1*nCell;
        //val = nCell;
        val = nSurf;
        //Brick b(bricks[0]); Coord c(b.ijk_mx); c -= b.ijk_mn; val = (c.i+1)*(c.j+1)*(c.k+1);
        //printf("%d %d %d %d\n",val,nRot,nCell,nSurf);
        return true;
    }
    return false;
}

void BrickRot::removeGameSym()
{
    // find(sym(x)) != x, if brick x is asymmetric;

    int i(0), j;
    Brick b;

    while(i<bricks.size())
    {
        b = bricks[i]; b.rotII(); j = find(b.asBlock()); if(j>=0) replace(j);
        b = bricks[i]; b.rotJJ(); j = find(b.asBlock()); if(j>=0) replace(j);
        b = bricks[i]; b.rotKK(); j = find(b.asBlock()); if(j>=0) replace(j);
        ++i;
    }
}

int BrickRot::find(const Block& b) const
{
    // linear search
    
    int i;

    for(i=0;i<bricks.size();++i) if(blocks[i]==b) return i;

    return -1;
}

void BrickRot::replace(int i)
{
    // replace member i by member mx (and remove mx)

    int mx(bricks.size()-1);

    nMove -= nShift[i];
    if(i!=mx)
    {
        bricks[i] = bricks[mx];
        blocks[i] = blocks[mx];
        nijk[i] = nijk[mx];
        nShift[i] = nShift[mx];
    }
    bricks.pop_back();
    blocks.pop_back();

    nijk.pop_back();
    nShift.pop_back();
}
////////////////////////////////////////////////////////////////////////////////
