#ifndef CTPUZZLE_BRICK__
#define CTPUZZLE_BRICK__
////////////////////////////////////////////////////////////////////////////////
#ifdef WIN32 // WIN32
#pragma warning( disable : 4786 ) // mangled names bigger than 255 characters
#endif //  END IFDEF WIN32

#include "timer.h"
using namespace utility;

#include <memory.h>
#include <math.h>
#include <vector>
#include <algorithm>
using namespace std;
////////////////////////////////////////////////////////////////////////////////
typedef unsigned int uint;
enum
{
    NI=5, NJ=4, NK=3, NB=12, NB0=4, MXBC=6, CELLS=NI*NJ*NK
};
enum
{
    CIBITS=6,            // #bits for a cell-index (0..NI*NJ*NK-1)
    UINTCELLS=30,        // #cells packed in an uint
    CNSBITS=8,           // #bits reserved for a cells neighbour-state
    UINTCNS=4,           // #cell-neighbour-states packed in an uint
    BCBITS=3,            // #bits for a brick-cell-index (0..MXBC-1)
    MASK_4=(1<<4)-1,     // bits 0..3 set to one
    MASK_6=(1<<6)-1      // bits 0..5 set to one
};
enum
{
    CNSTATES=(1<<CNSBITS),  // #cell-neighbour-states
    NSTATES=CELLS*CNSTATES, // #neighbour-states
    NSBITS=CIBITS+CNSBITS,  // #bits for a neighbour-state-index (0..NSTATES-1)
    CNSMASK=(1<<CNSBITS)-1, // all cell-neighbour-state-bits set to one
    NSMASK=(1<<NSBITS)-1,   // all neighbour-state-bits set to one
    BMASK=(1<<NB)-1,        // all brick-bits set to one
    BMASK0=(1<<NB0)-1,      // brick-bits 0..NB0-1 set to one
    BMASK1=BMASK&~BMASK0,   // brick-bits NB0..NB-1 set to one
    BCMASK=(1<<BCBITS)-1    // all brick-cell-bits set to one
};
enum
{
    // #uints to contain a cell-state (=ceil(CELLS/UINTCELLS))
    CSUINT= ((CELLS/UINTCELLS)*UINTCELLS==CELLS)?CELLS/UINTCELLS:CELLS/UINTCELLS+1,
    // #uints to contain a neighbour-state (=ceil(CELLS/UINTCNS))
    NSUINT= ((CELLS/UINTCNS)*UINTCNS==CELLS)?CELLS/UINTCNS:CELLS/UINTCNS+1,
    // #uints to contain a state (cells+neighbours)
    SUINT=CSUINT+NSUINT,
    // the same for bytes
    CSBYTE=CSUINT*sizeof(uint),
    NSBYTE=NSUINT*sizeof(uint),
    SBYTE=SUINT*sizeof(uint)
};
////////////////////////////////////////////////////////////////////////////////
// large neighbourhood
enum
{
    CNSBITSL=14,    // #bits reserved for a cells neighbour-state
    UINTCNSL=2      // #cell-neighbour-states packed in an uint
};
enum
{
    CNSTATESL=(1<<CNSBITSL),  // #cell-neighbour-states
    CNSMASKL=CNSTATESL-1,     // bit mask for cell-neighbour-states
    NSTATESL=CELLS*CNSTATESL  // #neighbour-states
};
enum
{
    // #uints to contain a neighbour-state
    NSUINTL= ((CELLS/UINTCNSL)*UINTCNSL==CELLS)?CELLS/UINTCNSL:CELLS/UINTCNSL+1,
    // the same for bytes
    NSBYTEL=NSUINTL*sizeof(uint)
};
////////////////////////////////////////////////////////////////////////////////
// 3D coordinate vector
struct Coord
{
    Coord() {}
    Coord(int ii,int jj,int kk): i(ii), j(jj), k(kk) {}
    Coord& operator +=(const Coord& rhs); 
    Coord& operator -=(const Coord& rhs); 
    void rotIJK() { int h(i); i = j; j = k; k = h; }
    void rotK() { int h(i); i = -j; j = h; }
    void rotII() { j = -j; k = -k; }
    void rotJJ() { i = -i; k = -k; }
    void rotKK() { i = -i; j = -j; }

    int i, j, k;
};
////////////////////////////////////////////////////////////////////////////////
// mapping coordinate to cell-index
inline uint cellIndex(int i,int j,int k) { return k+NK*(j+NJ*i); }
inline int cellIndex(const Coord& c) { return cellIndex(c.i,c.j,c.k); }
////////////////////////////////////////////////////////////////////////////////
// set and unset bits in bitmask-arrays
inline void bmSet(uint* p,const uint* q,int n)
{
    uint *pe=p+n; while(p!=pe) *p++|=*q++;
}
// maybe unnecessary specialisations
inline void bmSetState(uint* p,const uint* q)
{
    uint *pe=p+SUINT; while(p!=pe) *p++|=*q++;
}
inline void bmUnSetState(uint* p,const uint* q)
{
    uint *pe=p+SUINT; while(p!=pe) *p++&=~*q++;
}
////////////////////////////////////////////////////////////////////////////////
// more comfortable representation of a cell-state than a raw bitmask-array
struct Block
{
    uint hit(int i,int j,int k) const;
    uint hit(const Coord& c) const { return hit(c.i,c.j,c.k); }
    void set(const Coord& c);
    void unSet(const Coord& c);
    bool operator==(const Block& b) const;
    void clear();
    void dump() const;
    uint m[CSUINT];
};
////////////////////////////////////////////////////////////////////////////////
// for easier manipulation the cells of a Brick are given by indicees not Coords
struct Brick
{
    void addCell(int i,int j,int k);
    Block asBlock() const;
    void shift(const Coord& dijk);
    void shiftToOrigin();
    void swapBoundingBox();
    bool rangeCheck() const;
    float minCellIndex() const;
    void rotIJK();
    void rotK();
    void rotII();
    void rotJJ();
    void rotKK();

    vector<Coord> cells;
    Coord ijk_mn, ijk_mx;
};
////////////////////////////////////////////////////////////////////////////////
// get all possible orientations and the shift-ranges of a Brick
struct BrickRot
{
    void init(int id,const Brick& b);
    void removeGameSym();
    bool qualify(float& val) const;
    int find(const Block& bx) const;
    void replace(int i);
    bool operator<(const BrickRot& rhs) const { return nMove<rhs.nMove; }

    vector<Brick> bricks;
    vector<Block> blocks;
    vector<Coord> nijk;
    vector<int> nShift;
    int nCell, nRot, nMove, nSurf, brickId;
};
////////////////////////////////////////////////////////////////////////////////
#endif
